/*

    PGUI - PhOSCo GUI Testing
    Copyright (C) 2000 Mike Calder-Smith, mike@gmot.demon.co.uk
    Guillemot Design Ltd - www.PhOSCo.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.net.*;
import java.util.*;
import java.io.*;
import java.text.*;
import javax.swing.*;
import com.sun.image.codec.jpeg.*;

/**
<h1>PhOSCo GUI Test Robot</h1>
<pre>

    PGUI - PhOSCo GUI Testing
    Copyright (C) 2000 Mike Calder-Smith, mike@gmot.demon.co.uk
    Guillemot Design Ltd - www.PhOSCo.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


</pre>


<p>The PhOSCo GUI Test Robot is a very easy to use tool for testing the
Graphical User Interface (GUI) behaviour of Java applications.  It
consists of two programs and one helper class:
<ul>
<li>The PhOSCo GUI Capture tool <b><i>PGUICapture</b></i>,
<li>The PhOSCo GUI Robot program <b><i>PGUIRobot</b></i>, and
<li>The PhOSCo GUI Robot Slave helper class
<b><i>PGUIRobotSlave.class</b></i>.
</ul>

<p>For documentation, see the <A HREF="PGUIUser.html">PGUI user
documentation </A>.

<p>This is a very initial release with minimal testing, but does work.
Please let the author know of any problems.  Improvements and
adaptations welcomed.

<h2>PhOSCo Graphical User Interface Tester Capture Tool</h2>

<p>This code requires Java 1.3 to compile, and the availability of the
com.sun.image.codec.jpeg extensions from the Sun JDK 1.3 or JRE.

<p>This routine is used in testing the GUI interface of Java programs.
It is the utility which captures key and mouse events to file.
<p>Events are recorded to the file named in the runtime parameter
supplied at program start.  The program itself prompts for the tested
process invocation string.
<p>The Standard Output and Standard Error streams of the tested process
are redirected to files.
*/
public class PGUICapture extends Component {

  /**
  Version
  */
  static String version = "1.0";

  /**
  Copyright
  */
  static String copyr = "Copyright (c) 2000 Guillemot Design Ltd. See 'PhOSCo Open Source Licence' document.";

  /**
  Test Code
  */
  static final int COMMENT = 0;

  /**
  Test Code
  */
  static final int DELAY = 1;

  /**
  Test Code
  */
  static final int MOUSECLICK = 2;

  /**
  Test Code
  */
  static final int MOUSEDBLCLICK = 3;

  /**
  Test Code
  */
  static final int MOUSEMOVE = 4;

  /**
  Test Code
  */
  static final int SETAUTODELAY = 5;

  /**
  Test Code
  */
  static final int SETMODEKEY = 6;

  /**
  Test Code
  */
  static final int TYPEKEY = 7;

  /**
  Test Code
  */
  static final int UNSETMODEKEY = 8;

  /**
  Set meta key standard message
  */
  static String SET = "SET ";

  /**
  UnSet meta key standard message
  */
  static String UNS = "UNS ";

  /**
  Drag Mouse standard message
  */
  static String DM = "DM ";

  /**
  Move Mouse standard message
  */
  static String MM = "MM ";

  /**
  Test parameter - auto delay interval
  */
  int autodelayint;

  /**
  Test parameter - mouse button mask
  */
  int buttonmask;

  /**
  Test parameter - Key code
  */
  int keycode;

  /**
  Test parameter - delay interval
  */
  int sleeptime;

  /**
  Test parameter - mouse x coordinate
  */
  private static int mouse_x;

  /**
  Test parameter - mouse y coordinate
  */
  private static int mouse_y;

  /**
  Shift lock
  */
  boolean shiftlock;

  /**
  Log file
  */
  protected static PrintStream log;  // only one log

  /**
  The tested process
  */
  static Process   thisProc;

  /**
  Tested Process Screen X Origin
  */
  private int sox;

  /**
  Tested Process Screen Y Origin
  */
  private int soy;

  /**
  Tested Process Screen Width
  */
  private int swidth;

  /**
  Tested Process Screen Height
  */
  private int sheight;

  /**
  Verbose Mouse flag (if true, lots of Mouse moves; otherwise compacted).
  */
  boolean verboseMouse;

  /**
  Mouse move eaten flag
  */
  boolean mouseMoveEaten;

  /**
  Instrumentation Datagram Socket
  */
  DatagramSocket ds;

  /**
  Instrumentation Datagram Packet Buffer
  */
  byte[] buf;

  /**
  Instrumentation Datagram Packet
  */
  DatagramPacket dpckt;

  /**
  Log message written by logPrint()
  */
  String lastLogMsg;

  /**
  Last Drag message read by logPrint()
  */
  String lastDragMsg;

  /**
  Flag for Mouse Drag in process.
  */
  boolean eatingDrags;

  /**
  Mouse Released Flag
  */
  boolean mouseReleased;

  /**
  Saved Button Mask
  */
  int savedMods;

  /**
  Error Output Stream produced by the tested process
  */
  BufferedInputStream processErrors;

  /**
  Standard Output Stream produced by the tested process
  */
  BufferedInputStream processOutput;

  /**
  Class to write an output stream of the tested process to a file.
  */
  private final class WriteProcessOutput implements Runnable {

    BufferedInputStream outp;
    FileOutputStream output;
    boolean keepgoing;

    /**
    Constructor.
    @param BufferedInputStream Tested Process Output Stream
    @param String Output file name
    */
    public WriteProcessOutput( BufferedInputStream b1, String filename) {
      outp = b1;
      keepgoing = true;
      try {
        output = new FileOutputStream( filename, true );
      } catch( Exception e) {
        System.out.println("Process output file creation failure on file: " +
                            filename + " Exception: " + e );
        keepgoing = false;
      }

    }

    public void run() {
      int i;

      if (!keepgoing) return;

      try {
        i = 0;
        while(i != -1) {
          if (!keepgoing) break;
          i = outp.read();
          if (i==10) output.write(13);
          output.write(i);
        }
        output.close();
      } catch( Exception e) {
        System.out.println("Process output stream failure: " + e );
      }

      return;
    }

  }

  /**
  Mainline.
  @param String[] Runtime arguments:
  <ul>
  <li>Test script file name
  </ul>
  */
  public static void main (String args[]) {
    int i;
    int ok;
    String    testfilename = "";
    PGUICapture thisPGUI;
    int nargs;


    for (i=0; i<args.length; i++) {
      if (i==0) testfilename = args[0];
    }

    System.out.println("\r\nPhOSCo GUI Robot Capture Utility");
    System.out.println("\r\n\r\nCopyright (c) 2000 Guillemot Design Ltd. See www.phosco.com for details.\r\n");
    System.out.println("\r\nLicenced materials subject to the PhOSCo General Licence");

    ok = 1;
    if (testfilename == null)  {
      ok=0;
    } else if (testfilename.length()==0)  {
      ok=0;
    }

    if ( ok == 0 ) {
      System.out.println("\r\nInvoke program by: java PGUICapture testfilename " );
      System.out.println("\r\n   testfilename = " );
      System.exit(-1);
    }

    try {
      log = new PrintStream( new FileOutputStream(testfilename), true );
    } catch(IOException e) {
      System.out.println("Unable to open log file - logging inoperative");
    }


    thisPGUI = new PGUICapture();

    thisPGUI.getData();

    log.close();

    System.exit(0);
  }

  /**
  Constructor
  */
  public PGUICapture() {
    int i;
    String cmdstring;
    String s;
    Runtime   thisRT;

    // Invoke program.

    cmdstring = (String)JOptionPane.showInputDialog(null,
                "Invocation String", "Program under Test",
                JOptionPane.PLAIN_MESSAGE, null, null, "" );

    logPrint( "* PGUICapture Test Script File - Utility copyright (c) 2000 M A Calder-Smith" );
    logPrint( "* File created " +
          DateFormat.getDateTimeInstance().format(new java.util.Date(System.currentTimeMillis())) );
    logPrint( "*" );

    try {
      thisRT = Runtime.getRuntime();
      thisProc = thisRT.exec( cmdstring );

      logPrint( "INVOKED " + cmdstring );

      processErrors = new BufferedInputStream( thisProc.getErrorStream() );
      WriteProcessOutput wpe = new WriteProcessOutput( processErrors, "PGUICSTE.DAT" );
      new Thread(wpe).start();

      processOutput = new BufferedInputStream( thisProc.getInputStream() );
      WriteProcessOutput wpo = new WriteProcessOutput( processOutput, "PGUICSTO.DAT" );
      new Thread(wpo).start();

    } catch( Exception e) {
      s = new String("System cannot run application "+ cmdstring + "; exception " + e );
      System.out.println( s );
      logPrint( s );
      s = new String("\n\nCheck spelling of command, correct path, full extension on .bat commands, ... \n\n" );
      System.out.println( s );
      logPrint( s );
      return;
    }

    mouseMoveEaten = false;

    //
    //  Create Datagram stuff and listen for contact; attempt to find
    //  tested process screen limits so we don't click the mouse outside
    //  and so lose focus.
    //

    try {

      // A docile tested process will create a datagram

      ds = new DatagramSocket(9877);    // Capture port.
      ds.setSoTimeout( 30000 );         // Wait n seconds for read
                                        // before failing
      buf = new byte[256];
      dpckt = new DatagramPacket( buf, 256 );
      ds.receive( dpckt );
      buf = dpckt.getData();
      s = new String(buf);
      if (s != null) {
        if (s.startsWith( "Are you" )) {
          byte[] sendbuf = new byte[256];
          DatagramPacket senddpckt = new DatagramPacket( sendbuf, 256, InetAddress.getLocalHost(), 9876 );
          s = new String("All agog");
          int l = s.length();
          System.arraycopy( s.getBytes(), 0, sendbuf, 0, l );
          senddpckt.setData( sendbuf );
          ds.send( senddpckt );
        }
      }

      // get limits datagram

      ds.receive( dpckt );
      buf = dpckt.getData();
      s = new String(buf);
      if (s != null) {
        StringTokenizer st = new StringTokenizer( s );
        i = st.countTokens();
        if (i == 5) {
          String parm = st.nextToken();
          if (parm.equals("USING") ) {
            sox     = Integer.parseInt( st.nextToken() );
            soy     = Integer.parseInt( st.nextToken() );
            swidth  = Integer.parseInt( st.nextToken() );
            sheight = Integer.parseInt( st.nextToken().trim() );
          }
          logPrint("USING " + sox + " " + soy + " " + swidth + " " + sheight );
          logPrint("AD 300");
        } else {
          throw new Exception("Incorrect number of parameters " + i + " in datagram " + s);
        }
      } else {
        throw new Exception("Null parameters in datagram");
      }

    } catch(Exception e) {

      // Curses.  Foiled again.

      String errs = new String( "Can\'t proceed - bad initial testee protocol; exception " + e);
      System.out.println( errs );
      logPrint( errs );
      System.exit(-2);

    }

    verboseMouse = false;

    return;
  }


  /**
  Write an action to the test log file.
  @param String Action to write.
  */
  private void logPrint( String s ) {

    if ((s.startsWith(SET)) || (s.startsWith(UNS))) {

      if (s.equals( lastLogMsg )) {

        // log.println( "* Duplicate ignored " + s );

        return;       // Ignore duplicate messages
      }

    } else if (!verboseMouse) {

      if (s.startsWith(MM)) {

        if (eatingDrags) {                        // Ignore Mouse Moves
          // log.println( "* Eating drags " + s );
          return;                                 // after drag starts
        }

      } else if (s.startsWith(DM)) {

        eatingDrags = true;                       // Flag Dragging
        lastDragMsg = s;                          // Remember drag msg
        return;

      } else if (s.startsWith("ENDDRAG")) {

        eatingDrags = false;                      // Clear flag and
        log.println( lastDragMsg );               // write last Drag msg
        return;

      }

    }

    lastLogMsg = s;
    log.println(s);

    return;
  }

  /**
  Listen for Datagrams from the tested process, and send to the
  appropriate handler.
  */
  private void getData() {
    String s;
    int i;
    long now;
    long then;
    long interval;

    try {
      ds.setSoTimeout( 0 );         // Disable socket timeout
    } catch( Exception e ) {
      s = new String("Unable to disable datagram socket timeout. Failing with exception " + e);
      logPrint( s );
      System.out.println( s );
      return;
    }

    then = System.currentTimeMillis();
    while(true) {
      String err1 = "Error on datagram packet receive. Exception ";
      String err2 = "Unrecognised Datagram ";

      try {

        ds.receive( dpckt );
        buf = dpckt.getData();
        s = new String(buf);

        if (s.startsWith("!!EOP")) {
          logPrint( "TERMINATE");
          return;
        }

        if ( s.startsWith("KEY") ) {

          handleKeyEvent( s.substring(4) );

        } else if ( s.startsWith("* ") ) {

          // Marker Comment direct from Slave

          logPrint( s );

        } else if ( s.startsWith("MOUSE") ) {

          handleMouseEvent( s.substring(6) );

        } else if ( (s.startsWith("SC")) ||
                    (s.startsWith("FC")) ||
                    (s.startsWith("PI")) ) {

          // Dialog Command direct from Slave

          logPrint( s );

        } else if ( s.startsWith("USING") ) {

          // Screen size change

          StringTokenizer st = new StringTokenizer( s );
          i = st.countTokens();
          if (i == 5) {
            String parm = st.nextToken();
            sox     = Integer.parseInt( st.nextToken() );
            soy     = Integer.parseInt( st.nextToken() );
            swidth  = Integer.parseInt( st.nextToken() );
            sheight = Integer.parseInt( st.nextToken().trim() );
          }
          logPrint("USING " + sox + " " + soy + " " + swidth + " " + sheight );

        } else {

          s = err2 + s;
          logPrint( s );
          System.out.println( s );

        }

        now = System.currentTimeMillis();
        interval = now - then;
        then = now;
        if (interval > 100000L) {
          s = (String)JOptionPane.showInputDialog(null,
                "Should test program be terminated?", "Wait time exceeded",
                JOptionPane.PLAIN_MESSAGE, null, null, "" );
          if (s == null) continue;
          if (s.toUpperCase().startsWith("Y")) {
            logPrint( "TERMINATE" );
            return;
          }
        }


      } catch( Exception e ) {

        s = err1 + e;
        logPrint( s );
        System.out.println( s );
        logPrint( "TERMINATE - exception " + e );
        return;

      }

    }


  }

  /**
  Handle Datagram message about Key event
  @param String Datagram message
  */
  private void handleKeyEvent( String dmsg ) {
    int kc;

    kc = getKeyCode(dmsg);
    if (kc<0) {
      System.out.println("Bad key code in " + dmsg);
      return;
    }

    //
    // "Normal" Key events are output as "KEY n" where n is the keycode.
    // If the key is a meta key (Alt, Ctrl, Shift...) then "SET n" and
    // "UNS n" are used instead.
    //

    // To compact the number of messages, we first check for "Meta"
    // keys.  Then we "type" a key for every  "RELEASED" event or
    // "TYPED" event ( these are pre-filtered by RobotSlave).

    // Meta Keys

    if ((kc == KeyEvent.VK_ALT) ||
        (kc == KeyEvent.VK_CONTROL) ||
        (kc == KeyEvent.VK_META) ||
        (kc == KeyEvent.VK_SHIFT) ) {

      if (dmsg.startsWith("PRESS")) {
        logPrint( SET + kc + " " + KeyEvent.getKeyText(kc) );
      } else if (dmsg.startsWith("RELEA")) {
        logPrint( UNS + kc + " " + KeyEvent.getKeyText(kc) );
      }

    }

    if ((dmsg.startsWith("RELEA"))  ||
         (dmsg.startsWith("TYPED"))) {

      if ( kc < 10) {
        logPrint( "KEY VK" + kc + " " + KeyEvent.getKeyText(kc) );
      } else {
        logPrint( "KEY " + kc + " " + KeyEvent.getKeyText(kc) );
      }

    }

  }

  /**
  Get Keycode from Datagram message
  @param String Datagram message
  @return int Keycode
  */
  private int getKeyCode(String dmsg) {
    int start;
    int end;
    int kc;
    String s;

    start = dmsg.indexOf("keyCode=") + 8;
    end = dmsg.indexOf(',', start);
    s = dmsg.substring( start, end );
    try {
      kc = Integer.parseInt(s);
    } catch(Exception e) {
      kc = -1;
    }
    return kc;
  }

  /**
  Handle Datagram message about Mouse event
  @param String Datagram message
  */
  private void handleMouseEvent( String dmsg ) {
    Point p;
    int mods;
    int clicks;
    String s;

    //
    // Mouse events are output as "CM m" for Click Mouse with m as the
    // button modifier, "DCM m" for double-click, "MM x y" for Move
    // Mouse to x, y co-ordinate.
    //
    // There can be a lot of Mouse Move events.  Normally we will
    // compress these and only output a Mouse Move before a click or
    // doubleclick.  Where intemediate moves are required, verboseMouse
    // mode can be switched on.
    //

    // Ignore "EXITED" events; ignore "PRESSED" and "RELEASED" as
    // "CLICKED" gives us what we want.
    // (Except RELEASED when dragging.)

    mods = getMods(dmsg);
    clicks = getClickCount(dmsg);

    if ( (dmsg.startsWith("MOVE")) || (dmsg.startsWith("ENTER")) ) {
      // Treat "ENTERED" as a "MOVE".

      if (mouseReleased) {
        logPrint( MM + mouse_x + " " + mouse_y );
        logPrint( "CM " + savedMods);
      }
      mouseReleased = false;

      p = getPoint(dmsg);
      if (p != null) {
        mouse_x = (int)p.getX();
        mouse_y = (int)p.getY();

        if (verboseMouse) {
          logPrint( MM + mouse_x + " " + mouse_y );
          mouseMoveEaten = false;
        } else {
          mouseMoveEaten = true;
        }
      } else {
        System.out.println("Bad coordinate in " + dmsg);
      }

    } else if (dmsg.startsWith("CLICKED")) {

      if (!verboseMouse) {
        logPrint( MM + mouse_x + " " + mouse_y );
        mouseMoveEaten = false;
      }
      mouseReleased = false;

      if (clicks == 1) {

        logPrint( "CM " + mods);

      } else if (clicks == 2) {

        logPrint( "DCM " + mods);

      }

    } else if (dmsg.startsWith("DRAGGED")) {

      if (!verboseMouse) {
        logPrint( MM + mouse_x + " " + mouse_y );
        mouseMoveEaten = false;
      }
      mouseReleased = false;

      p = getPoint(dmsg);
      if (p != null) {
        mouse_x = (int)p.getX();
        mouse_y = (int)p.getY();

        logPrint( DM + mouse_x + " " + mouse_y + " " + mods );
      } else {
        System.out.println("Bad coordinate in " + dmsg);
      }


    } else if ( (eatingDrags) && (dmsg.startsWith("RELEASED"))) {

      logPrint( "ENDDRAG" );

    } else if (dmsg.startsWith("RELEASED")) {

      mouseReleased = true;
      savedMods = mods;

    } else {

      if (mouseReleased) {
        logPrint( MM + mouse_x + " " + mouse_y );
        logPrint( "CM " + savedMods);
      }
      mouseReleased = false;

      if ((!verboseMouse) && (mouseMoveEaten)) {
        // logPrint( "* Moved here ... " + mouse_x + " " + mouse_y + " dmsg " + dmsg.substring(0, 20));
      }
      if (dmsg.startsWith("RELEASED")) {
        // logPrint( "* Released here ... " + mouse_x + " " + mouse_y );
      }

      // logPrint( "* Thrown away " + dmsg );

    }

  }

  /**
  Get Modifiers from Datagram message
  @param String Datagram message
  @return int Modifiers
  */
  private int getMods(String dmsg) {
    int start;
    int end;
    int m;
    String s;

    start = dmsg.indexOf("mods=") + 5;
    end = dmsg.indexOf(',', start);
    s = dmsg.substring( start, end );
    try {
      m = Integer.parseInt(s);
    } catch(Exception e) {
      m = -1;
    }

    // We want 1 for left button, 2 for right button.
    if (m == 16) m = 1;
    else if (m == 4) m = 2;

    return m;
  }

  /**
  Get Click Count from Datagram message
  @param String Datagram message
  @return int Number of clicks
  */
  private int getClickCount(String dmsg) {
    int start;
    int end;
    int clicks;
    String s;

    clicks = 0;
    start = dmsg.indexOf("clickCount=") + 11;
    if ( start > 10 ) {
      s = dmsg.substring( start ).trim();
      try {
        clicks = Integer.parseInt(s);
      } catch(Exception e) {
        clicks = -1;
      }
    }
    return clicks;
  }

  /**
  Get Coordinate point from Datagram message
  @param String Datagram message
  @return Point Coordinate
  */
  private Point getPoint(String dmsg) {
    int start;
    int end;
    int x;
    int y;
    Point p;
    String s;

    start = dmsg.indexOf(",(") + 2;
    end = dmsg.indexOf(',', start);
    s = dmsg.substring( start, end );
    try {
      x = Integer.parseInt(s);
      start = end + 1;
      end = dmsg.indexOf("),", start );
      s = dmsg.substring( start, end );
      y = Integer.parseInt(s);
      p = new Point( x, y );
    } catch(Exception e) {
      p = null;
    }
    return p;
  }


}
